$LOAD_PATH.unshift File.expand_path('../../lib', __FILE__)
require 'discard'

require 'database_cleaner'
require 'with_model'

DatabaseCleaner.strategy = :transaction

RSpec.configure do |config|
  config.before :suite do
    ActiveRecord::Base.establish_connection :adapter => 'sqlite3', database: ':memory:'
  end

  config.before :each do
    DatabaseCleaner.start
  end
  config.after :each do
    DatabaseCleaner.clean
  end

  config.extend WithModel

  config.expect_with :rspec do |expectations|
    expectations.include_chain_clauses_in_custom_matcher_descriptions = true
  end

  config.mock_with :rspec do |mocks|
    mocks.verify_partial_doubles = true
  end

  config.shared_context_metadata_behavior = :apply_to_host_groups
  config.filter_run_when_matching :focus
  config.example_status_persistence_file_path = "spec/examples.txt"
  config.disable_monkey_patching!
  config.warnings = true

  if config.files_to_run.one?
    config.default_formatter = 'doc'
  end

  config.order = :random

  Kernel.srand config.seed
end
